#!/usr/bin/env python3
"""
Fix Zip Filenames
by Anonymous
Mar 5, 2015

zip, 7za, and python all don't support SJIS filenames. This script does.

It's easy to change this to work with other encodings. You're welcome.

An actual fix has been created as a modification to zipfile.py. Please use it 
instead.
"""
from __future__ import print_function
import zipfile
import struct
import os
import sys

def unzipAFileHere(filename):
	"""
	How difficult is it actually to unzip a file properly? Not difficult at 
	all. Here I do it in less than 100 lines of Python with some help of 
	course from a zip library.
	This function is not as secure as you might hope.
	"""
	f = open(filename, 'rb')
	file_size = f.seek(0, 2)
	f.seek(0)
	# Get the first 10k for the header.
	filenames = []
	
	filename_length = 0
	done = False
	pos = 0
	while pos < file_size:
		f.seek(pos, 0)
		data = f.read(10000)
		if len(data) < 35:
			break
		#end if
		if b'PK\x05\x06' == data[:4]:
			# This is the one at the very end.
			# End of directory, yay
			#print('pk56', pos)
			break
		elif b'PK\x01\x02' == data[:4]:
			# List of things at the end.
			# End of directory, yay
			#print('pk12', pos)
			break
			#done = True
		elif b'PK\x03\x04' != data[:4]:
			print("Warning: Missing ZIP header. Is this actually a zip file?")
			print('header:', data[:4])
			# FIXME: pyflakes found this weird code.
			f.seek(-30-filename_length, 1)
			data = f.read(10000)
			print('data:', data[:80])
			break
		#end if
		filename_length = struct.unpack('<I', data[0x1a:0x1e])[0]
		#print(data[30:30+filename_length])
		filenames.append(data[30:30+filename_length].decode('sjis'))

		compressed_size = struct.unpack('<I', data[18:22])[0]

		pos += compressed_size + filename_length + 30
		if done: break
	#loop

	b = zipfile.ZipFile(filename)
	c = b.infolist()
	for i, d in enumerate(c):
		x = b.open(d.filename)
		# Create the necessary directories.
		q = filenames[i].split('/')
		q.pop()
		dir_curr = ''
		for dn in q:
			if dn == '..':
				print("Error: .. is not allowed in the path.", 
					file=sys.stderr)
				continue
			#end if
			if not os.path.exists(dir_curr + dn):
				os.mkdir(dir_curr + dn)
			#end if
			dir_curr += dn + '/'
		#next dn
		print(filenames[i])
		if os.path.exists(filenames[i]):
			# Let this one go.
			# This is a difficult problem to deal with.
			continue
		#end if
		open(filenames[i], 'wb').write(x.read())
	#next d
#end def unzipAFileHere(filename)

def main():
	if len(sys.argv) < 2:
		print("Usage: fixzipnames1.py bad.zip")
		exit(1)
	#end if
	filename = sys.argv[1]
	unzipAFileHere(filename)
	
#end def main()

if __name__ == '__main__':
	main()
#end if
