#!/usr/bin/env python3
"""
Zlib Tattoo
by Javantea
Sept 26, 2015

After all these years I finally get my Zlib Tattoo!
Steganography with zlib without the work!

TODO:
Fix zlibPy

Long-term:
x264
png
gzip
tar.gz
lzma
other video formats
proper stego
renpy
.git

"""
from __future__ import print_function
import zlib

def zlibTattoo(tattoo, plain_filename='/usr/share/dict/words', plaintext_chop=None, compressor=None):
    """
    Create a zlib tattoo (steganography)
    """
    if compressor == None:
        compress = zlib.compress
        decompressobj = zlib.decompressobj
        errorClass = zlib.error
    else:
        compress, decompressobj, errorClass = compressor
    #end if
    if isinstance(tattoo, str):
        # This may throw an error but I prefer that to crashing later on.
        tattoo = tattoo.encode('utf-8')
    #end if
    
    b = compress(open(plain_filename, 'rb').read(plaintext_chop))
    #print('Compressed Plaintext length:', len(b))

    blen = len(b)
    lowBound = blen//10
    if lowBound < 2:
        # Almost certainly won't work...
        lowBound = 2
    #end if
    highBound = (blen*7)//8
    if blen - highBound < 5:
        # Almost certainly won't work...
        highBound = blen - 5
    #end if
    if highBound - lowBound < 5:
        # Almost certainly won't work...
        print("In order for this brute force method to work, we need a fairly" +
            " large plaintext. Maybe try a different plaintext.")
        return None
    #end if
    found = -1
    for plaintext_len in range(blen//10, (blen*7)//8):
        zo = decompressobj()
        # Decompress half of b giving you half of the dictionary.
        c = zo.decompress(b[:plaintext_len])
        # We can't use blank decompress, can we? I dunno...
        if c == b'': continue
        d = None
        try:
            d = zo.decompress(tattoo)
        except errorClass:
            pass
        #end try
        if d != None and d != '':
            found = plaintext_len
            print('full:', c, d)
            break
        #end if
    #next i
    if found > 0:
        return b[:plaintext_len] + tattoo
    #end if
    """
    Traceback (most recent call last):
    File "<stdin>", line 1, in <module>
    zlib.error: Error -3 while decompressing data: invalid distance too far back
    """

    # Now we need to brute force I guess.
    if len(b) < 20:
        print("In order for this brute force method to work, we need a fairly" +
            " large plaintext. Maybe try a different plaintext.")
        return None
    #end if
    maxlen = 0
    found = -1
    # Not nearly as expensive as you might think.
    # It only checks from maxlen to len(tattoo)-1 backwards, exiting when it 
    # found a good maxlen, thus only going through a few loops depending on the 
    # tattoo length and plaintext length. It bails when it finds a good match.
    for plaintext_len in range(lowBound, highBound):
        # reverse search allows us to find maxlen and break.
        # Also don't check for shorter matches than our maxlen.
        for i in range(len(tattoo)-1, maxlen, -1):
            zo = decompressobj()
            # Decompress part of b giving you part of the plaintext.
            c = zo.decompress(b[:plaintext_len])
            if c == b'': continue
            d = None
            try:
                # Add the tattoo to the compressed stream.
                d = zo.decompress(tattoo[:i])
            except errorClass:
                # If it fails, try the next one.
                pass
            #end try
            if d != None and d != '':
                # If it succeeds, tell everyone about it.
                #print(i, len(d))
                # i is guaranteed to be greater than maxlen.
                maxlen = i
                found = plaintext_len
                break
            #end if
        #next i
        # We won't find better than this.
        if maxlen >= len(tattoo)-1:
            break
        #end if
    #next plaintext_len
    if maxlen <= 0:
        print("Your tattoo won't go into the compression. Try another plaintext or" +
            " another ciphertext.")
        return None
    #end if
    #print("We were able to save part of your tattoo. %r" % tattoo[:maxlen])

    """
    1 6
    2 6
    3 9
    4 16
    5 17
    6 17
    7 24
    8 25
    9 25
    10 31
    11 37
    12 38
    13 38
    14 44
    15 44
    16 44
    """

    #"""
    zo = decompressobj()
    # Decompress half of b giving you half of the dictionary.
    c = zo.decompress(b[:found])
    d = zo.decompress(tattoo[:maxlen])
    print('part:', c, d)
    #"""
    return b[:found] + tattoo[:maxlen]
#end zlibTattoo(tattoo, plain_filename='/usr/share/dict/words', plaintext_chop=None, compressor=None)

def bzipTattoo(tattoo, plain_filename='/usr/share/dict/words', plaintext_chop=None):
    """
    Create a zlib tattoo (steganography)
    """
    import bz2
    compressor = (bz2.compress, bz2.BZ2Decompressor, OSError)
    return zlibTattoo(tattoo, plain_filename, plaintext_chop, compressor)
#end def bzipTattoo(tattoo, plain_filename='/usr/share/dict/words', plaintext_chop=None)

def lzmaTattoo(tattoo, plain_filename='/usr/share/dict/words', plaintext_chop=None):
    """
    Create a zlib tattoo (steganography)
    """
    import lzma
    compressor = (lzma.compress, lzma.LZMADecompressor, lzma.LZMAError)
    return zlibTattoo(tattoo, plain_filename, plaintext_chop, compressor)
#end def lzmaTattoo(tattoo, plain_filename='/usr/share/dict/words', plaintext_chop=None)

def main():
    tattoo = b'hi there hackers!'
    print("zlib")
    for i in range(10, 1000, 10):
        compressed = zlibTattoo(tattoo, '/usr/share/dict/words', i)
        if compressed:
            print("Zlib Tattoo!", i, repr(compressed))
        #end if
    #next i

    # BZip2 is not quite ready yet. Probably the block size needs to be 
    # tweaked.
    print("BZ2")
    for i in range(1000, 10000, 1000):
        compressed = bzipTattoo(tattoo, '/usr/share/dict/words', i)
        if compressed:
            print("Bzip2 Tattoo!", i, repr(compressed))
        #end if
    #next i

    # LZMA works as well as zlib.
    print("LZMA/XZ")
    for i in range(10, 1000, 10):
        compressed = lzmaTattoo(tattoo, '/usr/share/dict/words', i)
        if compressed:
            print("LZMA Tattoo!", i, repr(compressed))
        #end if
    #next i
#end def main()

if __name__ == '__main__':
    main()
#end if
